
#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



import logging
import requests

from datetime import datetime

from src.base.utils import SimpleUtil as SU



class CalendarFetcher:
    """
    Fetches the timeslots, playlists and playlist entries as JSON
    via the API endpoints of Steering and Tank.
    """
    config = None
    logging = None
    has_already_fetched = False
    fetched_timeslot_data = None

    # Config for API Endpoints
    steering_calendar_url = None
    tank_playlist_url = None
    tank_session = None
    tank_secret = None



    def __init__(self, config):
        """
        Constructor

        Args:
            config (AuraConfig):    Holds the engine configuration
        """
        self.config = config
        self.logger = logging.getLogger("AuraEngine")
        self.steering_calendar_url = self.config.get("api_steering_calendar")
        self.tank_playlist_url = self.config.get("api_tank_playlist")
        self.tank_session = self.config.get("api_tank_session")
        self.tank_secret = self.config.get("api_tank_secret")


    #
    #   PUBLIC METHODS
    #


    def fetch(self):
        """
        Retrieve all required data from the API.
        """
        return_data = []

        self.logger.debug("Fetching timeslots from STEERING")
        self.fetched_timeslot_data = self.fetch_timeslot_data()
        if not self.fetched_timeslot_data:
            self.logger.critical(SU.red("No timeslots fetched from API!"))
            return None

        self.logger.debug("Fetching playlists from TANK")
        self.fetch_playlists()

        try:
            for timeslot in self.fetched_timeslot_data:
                # Skip timeslot if no start or end is given
                if "start" not in timeslot:
                    self.logger.warning("No start of timeslot given. Skipping timeslot: " + str(timeslot))
                    timeslot = None
                if "end" not in timeslot:
                    self.logger.warning("No end of timeslot given. Skipping timeslot: " + str(timeslot))
                    timeslot = None
                if "playlist" not in timeslot \
                    and "show_fallback" not in timeslot \
                    and "schedule_fallback" not in timeslot \
                    and "station_fallback" not in timeslot:
                    
                    self.logger.warning("No playlist for timeslot given. Skipping timeslot: " + str(timeslot))
                    timeslot = None

                if timeslot:
                    return_data.append(timeslot)
        except TypeError:
            self.logger.error(SU.red("Nothing fetched ..."))
            self.fetched_timeslot_data = None
            return None

        return return_data



    #
    #   PRIVATE METHODS
    #


    def fetch_timeslot_data(self):
        """
        Fetches timeslot data from Steering.

        Returns:
            ([Timeslot]):   An array of timeslots
        """
        timeslot = None
        headers = { "content-type": "application/json" }
       
        try:
            self.logger.debug("Fetch timeslots from Steering API...")                
            response = requests.get(self.steering_calendar_url, data=None, headers=headers)
            if not response.status_code == 200:
                self.logger.critical(SU.red("HTTP Status: %s | Timeslots could not be fetched! Response: %s" % \
                    (str(response.status_code), response.text)))
                return None            
            timeslot = response.json()                
        except Exception as e:
            self.logger.critical(SU.red("Error while requesting timeslots from Steering!"), e)

        if not timeslot:
            self.logger.error(SU.red("Got no timeslot via Playout API (Steering)!"))
            return None

        return self.remove_unnecessary_data(timeslot)



    def fetch_playlists(self):
        """
        Fetches all playlists including fallback playlists for every timeslot.
        This method used the class member `fetched_timeslot_data`` to iterate
        over and extend timeslot data.
        """
        # store fetched entries => do not have to fetch playlist_id more than once
        fetched_entries=[]

        try:
            for timeslot in self.fetched_timeslot_data:

                # Get IDs of playlists
                playlist_id = self.get_playlist_id(timeslot, "playlist_id")
                schedule_fallback_id = self.get_playlist_id(timeslot, "schedule_fallback_id")
                show_fallback_id = self.get_playlist_id(timeslot, "show_fallback_id")
                station_fallback_id = self.get_playlist_id(timeslot, "station_fallback_id")

                # Retrieve playlist and the fallback playlists for every timeslot.
                # If a playlist (like station_fallback) is already fetched, it is not fetched again but reused
                timeslot["playlist"]          = self.fetch_playlist(playlist_id,          fetched_entries)
                timeslot["schedule_fallback"] = self.fetch_playlist(schedule_fallback_id, fetched_entries)
                timeslot["show_fallback"]     = self.fetch_playlist(show_fallback_id,     fetched_entries)
                timeslot["station_fallback"]  = self.fetch_playlist(station_fallback_id,  fetched_entries)

        except Exception as e:
            self.logger.error("Error while fetching playlists from API endpoints: " + str(e), e)



    def fetch_playlist(self, playlist_id, fetched_playlists):
        """
        Fetches the playlist for a given timeslot.

        Args:
            id_name (String):       The type of playlist to fetch (e.g. normal vs. fallback)
            fetched_playlists ([]): Previously fetched playlists to avoid re-fetching

        Returns:
            (Playlist):             Playlist of type `id_name`
        """
        if not playlist_id:
            return None
        playlist = None
        url = self.tank_playlist_url.replace("${ID}", playlist_id) 
        headers = {
            "Authorization": "Bearer %s:%s" % (self.tank_session, self.tank_secret), 
            "content-type": "application/json"
        }

        # If playlist is already fetched in this round, use the existing one
        for playlist in fetched_playlists:
            if playlist["id"] == playlist_id:
                self.logger.debug("Playlist #%s already fetched" % playlist_id)
                return playlist
                   
        try:
            self.logger.debug("Fetch playlist from Tank API...")             
            response = requests.get(url, data=None, headers=headers)
            if not response.status_code == 200:
                self.logger.critical(SU.red("HTTP Status: %s | Playlist #%s could not be fetched or is not available! Response: %s" % \
                    (str(response.status_code), str(playlist_id), response.text)))
                return None          
            playlist = response.json()                  
        except Exception as e:
            self.logger.critical(SU.red("Error while requesting playlist #%s from Tank" % str(playlist_id)), e)
            return None

        fetched_playlists.append(playlist)
        return playlist
 


    def get_playlist_id(self, timeslot, id_name):
        """
        Extracts the playlist ID for a given playlist (fallback) type.
        """
        playlist_id = str(timeslot[id_name])
        if not playlist_id or playlist_id == "None":
            self.logger.debug("No value defined for '%s' in timeslot '#%s'" % (id_name, timeslot["timeslot_id"]))
            return None
        
        return playlist_id


    def remove_unnecessary_data(self, timeslot):
        """
        Removes all timeslots which are not relevant for 
        further processing.
        """
        count_before = len(timeslot)
        timeslot = self.remove_data_more_than_24h_in_the_future(timeslot)
        timeslot = self.remove_data_in_the_past(timeslot)
        count_after = len(timeslot)

        self.logger.debug("Removed %d unnecessary timeslots from response. Entries left: %d" % ((count_before - count_after), count_after))
        return timeslot


    def remove_data_more_than_24h_in_the_future(self, timeslots):
        """ 
        Removes entries 24h in the future and 12 hours in the past.
        Note: This might influence resuming (in case of a crash)  
        single timeslots which are longer than 12 hours long.
        Think e.g. live broadcasts.
        """
        items = []
        now = SU.timestamp()
        now_plus_24hours = now + (12*60*60)
        now_minus_12hours = now - (12*60*60)

        for s in timeslots:
            start_time = datetime.strptime(s["start"], "%Y-%m-%dT%H:%M:%S")
            start_time = SU.timestamp(start_time)

            if start_time <= now_plus_24hours and start_time >= now_minus_12hours:
                items.append(s)

        return items


    def remove_data_in_the_past(self, timeslots):
        """
        Removes all timeslots from the past, except the one which is 
        currently playing.
        """
        items = []
        now = SU.timestamp()
        for s in timeslots:
            start_time = datetime.strptime(s["start"], "%Y-%m-%dT%H:%M:%S")
            start_time = SU.timestamp(start_time)
            end_time = datetime.strptime(s["end"], "%Y-%m-%dT%H:%M:%S")
            end_time = SU.timestamp(end_time)

            # Append all elements in the future
            if start_time >= now:
                items.append(s)
             # Append the one which is playing now
            elif start_time < now < end_time:
                items.append(s)

        return items
